/*
    i2c-MachZ.c - Linux kernel module for hardware monitoring
    Copyright (c) 2001 Hjalmar Hevers ,hhevers@ste-el.de>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

/*
    This driver supports the ACCESS.BUS-Device on the MachZ from ZFLinux
*/


#include <linux/version.h>
#include <linux/module.h>
#include <linux/pci.h>
#include <asm/io.h>
#include <asm/ioctl.h>
#include <linux/kernel.h>
#include <linux/stddef.h>
#include <linux/sched.h>
#include <linux/ioport.h>
#include <linux/init.h>
#include <linux/i2c.h>

/* MachZ-dependent Registers */
#define S_IO_BASEINDEXADR  0x2E
#define S_IO_BASEDATAADR   0x2F
#define S_IO_LOGICDEVICEOFFSET 0x07
#define S_IO_DEVICECONTROL 0x30
#define S_IO_DEVICEPORTADRESS  0x60
#define S_IO_DEVICEINTERRUPT   0x70
#define S_IO_DEVICEINTTYPE   0x71

#define S_IO_DEVENABLE  0x01
#define S_IO_DEVDISABLE  0
#define S_IO_SMBINTTYPE  0
#define S_IO_SMBINTERRUPT 0x07

#define SMB_LDN 0x08
#define SMB_BASE 0x200
#define SMB_DATAOFFSET 0
#define SMB_STATUSOFFSET 1
#define SMB_CONSTATOFFSET 2
#define SMB_CONTROL1OFFSET 3
#define SMB_OWNADROFFSET 4
#define SMB_CONTROL2OFFSET 5



/* PCI Address Constants */
#define SMBBA     0x020
#define SMBHSTCFG 0x040
#define SMBREV    0x008


/* Other settings */
#define MAX_TIMEOUT 50
#define  ENABLE_INT9 0
#define SMB_MAJOR 89
/* SMB_MachZ ioctl command constants */
#define SMB_MachZ_MAGIC          'p'
#define SMB_MachZ_SETADRESS       _IOW( SMB_MachZ_MAGIC,0,1)
#define SMB_MachZ_GETREG0         _IOR( SMB_MachZ_MAGIC,1,1)
#define SMB_MachZ_GETREG1         _IOR( SMB_MachZ_MAGIC,2,1)
#define SMB_MachZ_GETREG2         _IOR( SMB_MachZ_MAGIC,3,1)
#define SMB_MachZ_GETREG3         _IOR( SMB_MachZ_MAGIC,4,1)
#define SMB_MachZ_GETREG4         _IOR( SMB_MachZ_MAGIC,5,1)
#define SMB_MachZ_SETREG0         _IOR( SMB_MachZ_MAGIC,6,1)
#define SMB_MachZ_SETREG1         _IOR( SMB_MachZ_MAGIC,7,1)
#define SMB_MachZ_SETREG2         _IOR( SMB_MachZ_MAGIC,8,1)
#define SMB_MachZ_SETREG3         _IOR( SMB_MachZ_MAGIC,9,1)
#define SMB_MachZ_SETREG4         _IOR( SMB_MachZ_MAGIC,10,1)

long long i2c_lseek(struct file * file, long long offset, int origin) {	return -ESPIPE; };
int i2c_open(struct inode * inode, struct file * file);
int i2c_release (struct inode * inode, struct file * file);
ssize_t i2c_write(struct file * file,const char * buf, size_t count, loff_t *ppos);
ssize_t i2c_read(struct file * file,char * buf, size_t count, loff_t *ppos) ;
int i2c_ioctl(struct inode *inode, struct file *file,unsigned int cmd, unsigned long arg);

struct file_operations i2c_fops = {
	i2c_lseek,
	i2c_read,
	i2c_write,
	NULL,  			/* i2c_readdir	*/
	NULL,			/* i2c_select 	*/
	i2c_ioctl,
	NULL,			/* i2c_mmap 	*/
	i2c_open,
	NULL,			/* i2c_flush	*/
	i2c_release,
};

struct smb_dev {
  unsigned char adress;
  unsigned char status;
  unsigned char write;
  unsigned char timeout;
  unsigned long count;
  unsigned long maxcount;
  unsigned char data[32];
} smb_device = {0,}; 

/* insmod parameters */

/* If force_addr is set to anything different from 0, we forcibly enable
   the SMB_MachZ at the given address. VERY DANGEROUS! */
static int force_addr = 0;
MODULE_PARM(force_addr, "i");
MODULE_PARM_DESC(force_addr,
		 "Forcibly enable the SMB_MachZ at the given address. "
		 "EXTREMELY DANGEROUS!");
static int force_int = 0;
MODULE_PARM(force_int , "i");
MODULE_PARM_DESC(force_int ,
		 "Forcibly put the SMB_MachZ at the given Interrupt. "
		 "EXTREMELY DANGEROUS!");
#ifdef MODULE
static
#else
extern
#endif
int __init i2c_SMB_MachZ_init(void);
static int __init SMB_MachZ_cleanup(void);
static int SMB_MachZ_setup(void);
static void SMB_MachZ_do_pause(unsigned int amount);

#ifdef MODULE
extern int init_module(void);
extern int cleanup_module(void);
#endif				/* MODULE */

void SmbIRQhd (int, void * ,struct pt_regs *);

static int __initdata SMB_MachZ_initialized;
static unsigned short SMB_MachZ_smba = 0;
static unsigned short SMB_MachZ_int = 0;


/* Detect whether a SMB_MachZ can be found, and initialize it, where necessary.
   Note the differences between kernels with the old PCI BIOS interface and
   newer kernels with the real PCI interface. In compat.h some things are
   defined to make the transition easier. */
int SMB_MachZ_setup(void)
{
	int error_return = 0;
	unsigned char temp;
	unsigned long longtemp ;
	int inttemp;
	/* Note: we keep on searching until we have found 'function 3' */
	/* Determine the address of the SMBus areas */
	if (force_addr) {
		SMB_MachZ_smba = force_addr & 0xfff0;
	} else {
		SMB_MachZ_smba = SMB_BASE;
	}

	if (force_int) {
		SMB_MachZ_int = force_int & 0x0F;
	} else {
	  SMB_MachZ_int =  S_IO_SMBINTERRUPT & 0x0F;
	}
	if (check_region(SMB_MachZ_smba, 5)) {
	  printk
	    ("i2c-SMB_MachZ.o: SMB_MachZ_smb region 0x%x already in use!\n",
	     SMB_MachZ_smba);
	  error_return = -ENODEV;
	  goto END;
	}
	if( request_irq(SMB_MachZ_int,SmbIRQhd,0,"SMB",NULL) !=0){
	  printk
	    ("i2c-SMB_MachZ.o: SMB_MachZ_int 0x%x already in use! 0x%x \n",
	     SMB_MachZ_int,longtemp);
	  error_return = -ENODEV;
	  goto END;
	}
	/* We program the new address here. Just to make
	   sure, we disable the SMB_MachZ first. */
	outb_p(   S_IO_LOGICDEVICEOFFSET, S_IO_BASEINDEXADR   );
	outb_p(   SMB_LDN ,  S_IO_BASEDATAADR  );
	outb_p(   S_IO_DEVICECONTROL, S_IO_BASEINDEXADR );
	outb_p(   S_IO_DEVDISABLE,  S_IO_BASEDATAADR    );
	outb_p(   S_IO_DEVICEPORTADRESS+1, S_IO_BASEINDEXADR );
	outb_p(   (unsigned char)(SMB_MachZ_smba & 0x0FF), S_IO_BASEDATAADR );
	outb_p(   S_IO_DEVICEPORTADRESS, S_IO_BASEINDEXADR );
	outb_p(   (unsigned char)((SMB_MachZ_smba>>8) & 0x0FF), S_IO_BASEDATAADR );
	outb_p(   S_IO_DEVICEINTERRUPT, S_IO_BASEINDEXADR );
	outb_p(   (unsigned char)(SMB_MachZ_int&0x0FF), S_IO_BASEDATAADR );
	outb_p(   S_IO_DEVICEINTTYPE, S_IO_BASEINDEXADR );
	outb_p(   S_IO_SMBINTTYPE, S_IO_BASEDATAADR );
	outb_p(   S_IO_DEVICECONTROL, S_IO_BASEINDEXADR );
	outb_p(   S_IO_DEVENABLE, S_IO_BASEDATAADR);
	request_region(SMB_MachZ_smba, 5,"SMB");
 END:
	return error_return;
}


/* Internally used pause function */
void SMB_MachZ_do_pause(unsigned int amount)
{
	current->state = TASK_INTERRUPTIBLE;
	schedule_timeout(amount);
};

int __init i2c_SMB_MachZ_init(void)
{
	int res;
#ifdef DEBUG
/* PE- It might be good to make this a permanent part of the code! */
	if (SMB_MachZ_initialized) {
		printk
		    ("i2c-SMB_MachZ.o: Oops, SMB_MachZ_init called a second time!\n");
		return -EBUSY;
	}
#endif
	if ((res = SMB_MachZ_setup())) {
		printk
		    ("i2c-SMB_MachZ.o: SMB_MachZ not detected, module not inserted.\n");
		SMB_MachZ_cleanup();
		return res;
	}
	SMB_MachZ_initialized++;
	printk("i2c-SMB_MachZ.o: SMB_MachZ bus initialized\n");
	if (register_chrdev(SMB_MAJOR,"smb",&i2c_fops)) {
		printk("i2c: unable to get major %d for i2c bus\n",I2C_MAJOR);
	}
	smb_device.status=0;
	return 0;
}

int __init SMB_MachZ_cleanup(void)
{
	int res;

	outb_p(   S_IO_LOGICDEVICEOFFSET, S_IO_BASEINDEXADR );
	outb_p(     SMB_LDN,  S_IO_BASEDATAADR );
	outb_p(   S_IO_DEVICECONTROL, S_IO_BASEINDEXADR );
	outb_p(     S_IO_DEVDISABLE, S_IO_BASEDATAADR );
	if (SMB_MachZ_initialized >= 1) {
		release_region(SMB_MachZ_smba, 5);
		free_irq(SMB_MachZ_int,NULL);
		SMB_MachZ_initialized--;
	}
	unregister_chrdev(SMB_MAJOR,"smb");
	return 0;
}

EXPORT_NO_SYMBOLS;

#ifdef MODULE

MODULE_AUTHOR
    ("Frodo Looijaard <frodol@dds.nl>, Philip Edelbrock <phil@netroedge.com>, and Mark D. Studebaker <mdsxyz123@yahoo.com>");
MODULE_DESCRIPTION("SMB_MachZ SMBus driver");

int init_module(void)
{
	return i2c_SMB_MachZ_init();
}

int cleanup_module(void)
{
	return SMB_MachZ_cleanup();
}

#endif				/* MODULE */

void SmbIRQhd (int irq,void *dev_id ,struct pt_regs *reg_pnt){

};

int i2c_open(struct inode * inode, struct file * file){
  if(smb_device.status){
    printk("SMB-Device already open");
    return(-ENODEV);
  };
  smb_device.timeout=MAX_TIMEOUT;
  outb_p(((54<<1)+1), SMB_MachZ_smba+SMB_CONTROL2OFFSET   );
  while(smb_device.timeout!=0 && (inb_p(SMB_MachZ_smba+SMB_CONSTATOFFSET)&0x02)){
    current->state = TASK_INTERRUPTIBLE;
    schedule_timeout(1);
    smb_device.timeout--;
  }  ;
  if(smb_device.timeout==0){
     printk("SMB-BUS BUSY");
      return(-ENODEV);
  }
  MOD_INC_USE_COUNT;
  return (0);
};

int i2c_release (struct inode * inode, struct file * file){
  outb_p( (54<<1) ,SMB_MachZ_smba+SMB_CONTROL2OFFSET );//smb disable
  MOD_DEC_USE_COUNT;
  return(0);
};

ssize_t i2c_write(struct file * file,const char * buf, size_t count, loff_t *ppos){
  return(0);
};

ssize_t i2c_read(struct file * file,char * buf, size_t count, loff_t *ppos){
  return(0);
} ;

int i2c_ioctl(struct inode *inode, struct file *file,unsigned int cmd, unsigned long arg){
  
  switch(cmd){
  case 0x654 /* SMB_MachZ_SETADRESS */ :
    smb_device.adress=(unsigned char)(arg & 0xFF);
    return(0);
  case 0x655 /*  SMB_MachZ_GETREG0 */:
    return(inb_p(SMB_MachZ_smba));
  case  0x656 /*  SMB_MachZ_GETREG1 */:
    return(inb_p(SMB_MachZ_smba+1));
  case 0x657 /*  SMB_MachZ_GETREG2 */:
    return(inb_p(SMB_MachZ_smba+2));
  case 0x658 /*  SMB_MachZ_GETREG3 */:
    return(inb_p(SMB_MachZ_smba+3));
  case 0x659 /*  SMB_MachZ_GETREG4 */:
    return(inb_p(SMB_MachZ_smba+4));
  case 0x660 /*  SMB_MachZ_SETREG0 */:
    outb_p((unsigned char)(arg&0x0FF) ,SMB_MachZ_smba+0);
    return(0);
  case 0x661 /*  SMB_MachZ_SETREG1 */:
    outb_p((unsigned char)(arg&0x0FF) ,SMB_MachZ_smba+1);
    return(0);
  case 0x662 /*  SMB_MachZ_SETREG2 */:
    outb_p((unsigned char)(arg&0x0FF) ,SMB_MachZ_smba+2);
    return(0);
  case 0x663 /*  SMB_MachZ_SETREG3 */:
    outb_p((unsigned char)(arg&0x0FF) ,SMB_MachZ_smba+3);
    return(0);
  case 0x664 /*  SMB_MachZ_SETREG4 */:
    outb_p((unsigned char)(arg&0x0FF) ,SMB_MachZ_smba+4);
    return(0);
  };
};
